	  							      COMMENT ~
SOUND.ASM -- Sound Generation Routines

   From `BLUEBOOK of ASSEMBLY ROUTINES for the IBM PC & XT'
         by Christopher L. Morgan
         Copyright (C) 1984 by The Waite Group, Inc.
 
  SOUND.ASM contains a collection of assembly language routines for producing
    sound in MS/PCDOS 8088 assembly language, using MASM. These routines are
    FAR PROCEDURES.

   Contents:
   ---------
   DELAY	--  Delay for a specified time interval	
   FREQ		--  Convert from frequency to period
   GLISSNDO	--  Make a glissando (sliding tone)
   LINSCALE	--  Provide linear scaling
   PITCH	--  Convert from pitch number	
   PLAY		--  Play music from a table
   TONE		--  Make a tone
   TONE_INIT	--  Initialize speaker timer
   TONE_OFF	--  Turn off tone
   TONE_ON	--  Turn on tone
   TONE_SET	--  Set the tone on the speaker
   
 
  Overview:
  ---------
	The speaker is connected to the output of timer 2, which produces a 
  square wave whose frequency can be set using the TONE_SET routine.  TONE_SET
  assumes the timer has been properly initialized, as is done during normal
  boot-up.  TONE_INIT can also be used to initialize the timer.
	The frequency of the square wave from the timer is determined by the
  formula, f=F/n, where f is the frequency of the square wave, F is 1,193,182
  and n is a 16-bit integer that is input to the routine.  The constant, F, is
  exactly 1/3 the frequency of the NTSC subcarrier used for color encoding by
  the Color/Graphics Adapter.  F is derived by hardware from a main clock
  signal that runs at 14,317,800 Hertz and supplies the timing for most of the
  computer.
	TONE_ON and TONE_OFF provide control over whether the square wave
  reaches the speaker, thus turning the sound on or off.  There are actually
  two bits involved, one to control the connection between the clock signal
  and the timer, and one to control the connection between the timer and the
  speaker.  Both routines switch both connections simultaneously.  There is
  no provision to control the bits independently.
	The routine, DELAY, provides timing for the duration of musical notes
  in milliseconds.  The routine, FREQ, uses the formula, n=F/f, to compute
  the input parameter for TONE_SET from a given frequency.  Using this just
  before TONE_SET permits working directly with frequencies rather than clock
  cycles of the 1,193,182 Hertz clock.
	The routine, TONE, uses the other routines to produce tones of a
  given frequency and duration.  The frequency is input as a 16-bit integer
  and the duration is in milliseconds using the DELAY routine.
	The rest of the routines provide special sound effects.  SCALE converts
  numbers between 0 and 1 into numbers within the range of X1 and X2, which
  are any specified 16-bit integers.  This is used with RANDOM to produce
  pseudo-random numbers between 0 and 1.  The routine, WHITE, uses these 
  pseudo-random numbers along with SCALE to generate white noise.  The routine,
  GUN shows how to shape white noise into a machine gun sound.
	The routine, GLISSANDO, makes a tone that slides smoothly from one
  frequency, X1, to another, X2.  RED shows how to shape glissandos into red
  alert sirens.
	PITCH and PLAY are music routines.  PITCH converts the pitch number
  to a value useable by TONE_SET to set the frequency.  PLAY plays music from
  a "playlist" of notes.

__________________________ Descriptions of Routines __________________________

DELAY -- Delay for a specified number of milliseconds
-----------------------------------------------------
  Function: This routine provides a time delay for specifying the 
    duration of a sound, a wait, etc.  

  Input: Upon input, CX contains the no. of milliseconds to delay.

  Output: None

  Register used: CX is first saved and then restored.
_______________________________________________________________________________

FREQ -- Convert from Frequency to Period
----------------------------------------
  Function: This routine  converts from frequency to the number required
    by TONE_SET to set the frequency.  The routine evaluates the formula,
    N = F/f, where f is the frequency input to this routine, n is the number 
    output by this routine, and F is 1,193,182. In other words, this routine
    divides the specified frequency, f, into the 1,193,182 hertz clock
    frequency that drives the timer.  Use this routine just before TONE_SET.

  Input: Upon entry, the frequency is in CX.

  Output: Upon exit, F/f is in CX.

  Registers used: CX is modified, AX & DX are first saved and then restored.
_______________________________________________________________________________

GLISSNDO -- Make a Glissando
----------------------------
  Function: This routine makes a glissando; i.e., a sound that slides from
    one frequency to another.  The rate of change can be controlled.

  Input: Upon entry, the starting frequency is in BX, the ending frequency
    is in CX, and the control parameter for the rate increase is in DX.
    Increasing the value in DX slows down the rate of change.

  Output: To the speaker and timer only.

  Registers used: All registers are saved and then restored.

  Segments referenced: The data segment must contain the variables,
    F_START & F_END.

  Routines called: TONE_SET, TONE_ON, TONE_OFF, DELAY

  Special note: The speaker timer must already have been initialized.  If
    necessary, use TONE_INIT to initialize before calling this routine.
_______________________________________________________________________________

PITCH -- Convert from Pitch Number
----------------------------------
  Function: This routine converts from pitch number to the value required
    by TONE_SET to set the frequency. The pitch numbers refer to an 
    extended chromatic scale. The notes of this scale are numbered from
    0 to 95 with 12 notes/octave. 0 corresponds to a C at 16.35 hertz.

  Input: Upon entry the pitch number of the note is in AL.

  Output: Upon exit, the DX register has the proper value for TONE_SET.

  Registers used: DX has value; AX, BX, & CX are saved and then restored.

  Segments Referenced: The data segment must contain the
                        following pitch table:

	NOTES	DW	4186			;C
		DW	4435			;C#/D-
		DW	4699			;D
		DW	4978			;D#/E-
		DW	5274			;E
		DW	5588			;F
		DW	5920			;F#/G-
		DW	6272			;G
		DW	6645			;G#/A-
		DW	7040			;A
		DW	7459			;A#/B-
		DW	7902			;B

  Routines called: FREQ
_______________________________________________________________________________

LINSCALE -- Do Linear Scaling
-----------------------------
  Function: This routine performs a linear scaling, converting a fixed
    point number between 0 and 1 into an integer between X1 and X2, 
    where X1 and X2 are 16-bit integers.

  Input: Upon entry, CX has a binary fixed point number between 0 and 1.
    The binary point is to the left of the leftmost bit. X1 and X2 are
    variables stored in memory.

  Output: Upon exit, CX contains the 16-bit integer result.

  Registers used: CX is modified, AX & DX are first saved and then restored.

  Segments referenced: The data segment must contain the variables, X1 & X2.
_______________________________________________________________________________

PLAY -- Play Music
------------------
  Function: This program plays music.  It reads a binary "play" list
    that contains instructions to make the tune.  This list consists
    of a series of music instructions: Tempo, Note, Rest, and End.

    The syntax for the instructions are:

      Tempo Command
        first byte  = ASCII "T"
        second byte = tempo in whole notes/minute

      Note Command
        first byte  = ASCII "N"
        second byte = pitch number (integer 0-95)
        third byte  = length (8-bit binary fixed point - scale 1)
        fourth byte = style (8-bit binary fixed point - scale 0)

      Rest Command
        first byte  = ASCII "R"
        second byte = length (8-bit binary fixed point - scale 1)

      End Command
        first byte  = ASCII "X"

    The scaling is:
      scale 0 has the binary point to the left of the leftmost digit
      scale 1 has the binary point to the right of the leftmost bit

  Input: Upon entry, the address of the "play" list is in DS:SI

  Output: To the speaker and timer only

  Registers Used: Entry registers are saved and restored

  Segments Referenced: The data segment contains the "play" list and
                         the variables WHOLE, ACCOUNT, and RCOUNT

  Routines Called: PITCH, TONE_OFF, DELAY
_______________________________________________________________________________

TONE -- Make a Tone
-------------------
  Function: This routine makes a tone of a given frequency and given length.

  Input: Upon entry, the frequency is in CX and the length in no. of
    milliseconds is in DX.

  Output: To the speaker and timer only.

  Registers used: AX, CX, & DX are first saved and then restored.

  Segments referenced: The data segment must contain the variable, COUNT.

  Routines called: TONE_SET, TONE_ON, TONE_OFF, DELAY

  Special note: The speaker timer must already have been properly initialized.
    This should normally occur during boot-up.
_______________________________________________________________________________

TONE_INIT -- Initialize Speaker Timer
-------------------------------------
  Function: This routine initializes the part of the 8253 timer chip used
    by the speaker system.  Particularly, it sets up channel 2 of this 
    timer as a square wave generator. This routine does not select the 
    frequency, nor does it turn on the tone.  Use TONE_SET to select the
    frequency, TONE_ON to turn the tone on, and TONE_OFF to turn it off.

  Input: None

  Output: Only to the timer 2 of the speaker circuit.

  Registers used: AX is first saved and then restored.
_______________________________________________________________________________

TONE_OFF -- Turn off Tone
-------------------------
  Function: This routine turns of the timer and speaker.

  Input: None

  Output: To the timer and speaker only.

  Registers Used: AX register is first saved and then restored.
_______________________________________________________________________________

TONE_ON -- Turn on Tone
-----------------------
  Function: This routine turns on the timer and speaker to produce a tone.
    The frequency of the tone must have already been selected on the timer.
    you can use TONE_SET to set the frequency of the tone.

  Input: None

  Output: To the timer and speaker only.

  Register used: AX is first saved and then restored.
_______________________________________________________________________________

TONE_SET -- Set the Tone on the Speaker
---------------------------------------
  Function: This routine selects the frequency of the square wave tone to 
    the speaker.  The input to this routine is a 16-bit integer n which 
    determines the frequency, f, according to the following formula:
      f = F/n
    where F is 1,193,182, the frequency of a clock signal which feeds the
    timer.  The value, n, is the number of cycles of the clock signal per
    cycle of the resulting square wave.  This routine does not actually 
    turn on the tone.  Use TONE_ON to turn on the tone, and TONE_OFF to turn
    it off. This routine assumes the speaker timer has already been properly
    initialized.  This happens during normal boot-up, or you can use 
    TONE_INIT to initialize the timer.

  Input: Upon entry, the 16-bit integer, n, is in the CX register.

  Output: Only to the timer 2.

  Register used: AX is first saved and then restored.
_______________________________________________________________________________
 
 _____________________________ SOUND ROUTINES_________________________________
  It is best to include this data in the source code calling these routines,
    and then commenting out this next section.                                ~
    
DATAS	SEGMENT	PUBLIC
NOTES	DW	4186	;C
	DW	4435	;C#/D-
	DW	4699	;D
	DW	4978	;D#/E-
	DW	5274	;E
	DW	5588	;F
	DW	5920	;F#/G-
	DW	6272	;G
	DW	6645	;G#/A-
	DW	7040	;A
	DW	7459	;A#/B-
	DW	7902	;B
WHOLE	DW	0
F_START	DW	0
F_END	DW	0
DATAS	ENDS
;------------------------------------------------------------------------------
CODES	SEGMENT
PUBLIC	DELAY,FREQ,GLISSANDO,LINSCALE,PITCH,PLAY
PUBLIC  TONE,TONE_INIT,TONE_OFF,TONE_ON,TONE_SET
	ASSUME	 CS:CODES,DS:DATAS
;------------------------------------------------------------------------------
;Routine to set tone
;
TONE_INIT	PROC	FAR
	PUSH	AX				;Save register
;
;Define control bit field parameters for the timer chip
SC	=	2				;Use counter 2
RL	=	3				;Mode load period 1 byte/time
MODE	=	3				;Square wave generator
BCD	=	0				;Not BCD, use binary values
;
;Form control word
CNWORD	=	SC * 40H + RL * 10H + MODE * 2 + BCD
;
;Send control word to 8253 timer chip
	MOV	AL,CNWORD			;Select the above control word
	OUT	43H,AL				;Send it to the control port
	POP	AX				;Restore register
	RET
TONE_INIT	ENDP
;------------------------------------------------------------------------------
;Routine to select tone
;
TONE_SET	PROC	FAR
	PUSH	AX				;Save register
;
;Load the time period into the timer
	MOV	AL,CL				;Lower byte
	OUT	42H,AL				;Out to timer
	MOV	AL,CH				;Upper byte
	OUT	42H,AL				;Out to timer
	POP	AX				;Restore register
	RET
TONE_SET	ENDP
;------------------------------------------------------------------------------
;Routine to turn on tone
;
TONE_ON	PROC	FAR
	PUSH	AX				;Save register
;
;Turn speaker and timer on
	IN	AL,61H				;Get contents of system port B
	OR	AL,3				;Turn speaker and timer on
	OUT	61H,AL				;Send out new values to port B
	POP	AX				;Restore register
	RET
TONE_ON	ENDP
;------------------------------------------------------------------------------
;Routine to turn Tone off
;
TONE_OFF	PROC	FAR
	PUSH	AX				;Save register
;
;Turn off timer 2 and speaker
	IN	AL,61H				;Get port B again
	AND	AL,11111100B			;Turn off timer & speaker
	OUT	61H,AL				;Now do it
	POP	AX				;Restore register
	RET
TONE_OFF	ENDP
;------------------------------------------------------------------------------
;Routine to delay a specified number of milliseconds
;
DELAY	PROC	FAR
	PUSH	CX				;Save register
DELAY1:
	PUSH	CX				;Save counter
	MOV	CX,260				;Timing constant
DELAY2:
	LOOP	DELAY2				;Small loop
	POP	CX				;Restore counter
	LOOP	DELAY1				;Loop to count milliseconds
	POP	CX				;Restore register
	RET
DELAY	ENDP
;------------------------------------------------------------------------------
;Routine to convert from frequency to period
;
FREQ	PROC	FAR
	PUSH	DX				;Save registers
	PUSH	AX
	MOV	DX,12H				;Upper part of numerator
	MOV	AX,34DEH			;Lower part of numerator
	DIV	CX				;Divide by frequency
	MOV	CX,AX				;The quotient is the output
	POP	AX				;Restore registers
	POP	DX
	RET
FREQ	ENDP
;------------------------------------------------------------------------------
;Routine to make a tone
;
TONE	PROC	FAR
	PUSH	DX				;Save registers
	PUSH	CX		
	PUSH	AX
;
;Compute the frequency and set up the tone
	CALL	FREQ				;Convert the frequency
	CALL	TONE_SET			;Set up the tone
;
;Turn on the tone
	CALL	TONE_ON				;Turn it on
;
;Wait for proper delay
	MOV	CX,DX				;Get delay length
	CALL	DELAY				
;
;Turn off the tone
	CALL	TONE_OFF			;Turn it off
	POP	AX				;Restore registers
	POP	CX	
	POP	DX
	RET
TONE	ENDP
;------------------------------------------------------------------------------
;Routine to scale linearly
;
LINSCALE	PROC	FAR
	PUSH	DX				;Save registers
	PUSH	AX
;
;Compute width
	MOV	AX,F_END			;Get F_END
	SUB	AX,F_START			;Subtract F_START
;
;Multiply width by input parameter
	MUL	CX				;Multiply
	MOV	CX,DX				;Move top part of quotient
;						; into CX
;Add lower limit	
	ADD	CX,F_START			;Add F_START
	POP	AX				;Restore registers
	POP	DX
	RET
LINSCALE	ENDP
;------------------------------------------------------------------------------
;Routine to determine pitch
;
PITCH	PROC	FAR
	PUSH	CX				;Save registers
	PUSH	BX
	PUSH	AX
	MOV	AH,0				;Extend pitch no. to 16 bits
	MOV	CL,12				;Divisor of 12
	DIV	CL				;Divide
	MOV	DL,AL				;Quotient determines the octave
	MOV	AL,AH				;Remainder is the pitch within
	CBW					; 16-bit needed for look up
	SAL	AX,1				; 2 bytes/item
	MOV	BX,AX				; into BX
	MOV	CX,NOTES[BX]			;Look it up
	CALL    FREQ				;Convert the frequency
	XCHG	CX,DX				;Octave in CL, period in DX
	NEG	CL				;8 - octave = shift count
	ADD	CL,8	
	SAL	DX,CL
	POP	AX				;Restore registers
	POP	BX
	POP	CX
	RET
PITCH	ENDP
;------------------------------------------------------------------------------
;Routine to make glissando
;
GLISSANDO	PROC	FAR
	PUSH	SI				;Save registers
	PUSH	DX
	PUSH	CX
	PUSH	BX
	PUSH	AX
	MOV	F_START,BX			;FROM limit of frequencies
	MOV	F_END,CX			;  TO limit of frequencies
	CALL	TONE_ON				;Turn on tone
;
;Set up the loop parameters
	MOV	SI,1				;Increment for loop
	CMP	BX,CX				;Up or down?
	JLE	GLISS1				;Skip if up
	NEG	SI				;Decrement freq in the loop
GLISS1:
	MOV	CX,BX				;Get the frequency
	CALL	FREQ				;Convert to clock cycles
	CALL	TONE_SET			;Set the tone
	MOV	CX,DX				;Delay parameter > slows slide
GLISS2:
	LOOP	GLISS2
	CMP	BX,F_END			;Check if done
	JE	GLISS3				;If so, go
	ADD	BX,SI				;Else update the frequency
	JMP	GLISS1
;
;Turn off the tone
GLISS3:
	CALL	TONE_OFF			;Turn it off
	POP	AX				;Restore registers
	POP	BX
	POP	CX
	POP	DX
	POP	SI
	RET
GLISSANDO	ENDP	
;------------------------------------------------------------------------------
;Routine to play music
;
PLAY	PROC	FAR
	PUSH	DS				;Save registers
	PUSH	SI
	PUSH	DX
	PUSH	CX
	PUSH	BX
	PUSH	AX
;
;Command pointer is in SI
	MOV	WHOLE,2000			;Whole note = 2000 milliseconds
	CLD					;Forward direction
;
;Main loop starts here
CHEK_END:					;Get command code and go
						; through the cases
	LODSB					;Get the byte
;
;End command
	CMP	AL,'X'				;Is it End command?
	JNE	CHEK_TEMPO
	JMP	PLAY_XIT
;
;Tempo command
CHEK_TEMPO:
	CMP	AL,'T'				;Is it Tempo command?
	JNE	CHEK_NOTE
	LODSB					;Get the tempo
	MOV	CL,AL				;Set in CX
	MOV	CH,0			
	MOV	AX,60000			;No. of milliseconds/minute
	MOV	DX,0				;Clear upper part
	DIV	CX				;Divide into time
	MOV	WHOLE,AX			;No. of milliseconds/whole note
	JMP	CHEK_END			;Back for more
;
;Note command
CHEK_NOTE:
	CMP	AL,'N'				;Is it Note command?
	JNE	CHEK_REST
	LODSB					;Get the pitch
	CALL	PITCH				;Convert
	MOV	CX,DX				; and move result into CX
	CALL	TONE_SET			;Set the frequency
	CALL	TONE_ON				;Turn on the tone
	MOV	CX,WHOLE			;No. of milliseconds/whole note
	LODSB					;Get the duration
	MOV	AH,AL				;Set up duration as multiplier
	MOV	AL,0
	SAL	CX,1				;Scale factor 1
	MUL	CX				;Multiply
	MOV	CX,DX				;Total count for the note
	LODSB					;Get style
	MOV	AH,AL				;Set up style as multiplier
	MOV	AL,0				
	MUL	CX				;Multiply by style
	MOV	F_START,DX			;Store count for note
	SUB	CX,DX				;Count for rest
	MOV	F_END,CX			;Store count for rest
	MOV	CX,F_START			;Audible part of note
	CALL	DELAY				;Delay
	CALL 	TONE_OFF			;Turn off the tone
	MOV	CX,F_END			;Inaudible part of tone
	CALL	DELAY				;Delay
	JMP	CHEK_END			;Back for more
;
;Rest command
CHEK_REST:
	CMP	AL,'R'				;Is it Rest command?
	JNE	PLAY_XIT
	MOV	CX,WHOLE			;No. of milliseconds/whole note
	LODSB					;Get the duration
	MOV	AH,AL				;Set up duration as multiplier
	MOV	AL,0				
	SAL	CX,1				;Scale factor of 1
	MUL	CX				;Multiply
	MOV	CX,DX				;Total count
	CALL	DELAY				;Delay
	JMP	CHEK_END			;Back for more
;
;Anything else end it
PLAY_XIT:
	POP	AX				;Restore registers
	POP	BX
	POP	CX
	POP	DX
	POP	SI
	POP	DS
	RET
PLAY	ENDP
;-----------------------------------------------------------------------------
CODES	ENDS
;
	END
;_____________________________________________________________________________
;>>>>>Physical EOF SOUND.ASM<<<<<	
